@@ -1,7 +1,7 @@
 /*
  * framecpp channel demultiplexor
  *
- * Copyright (C) 2011--2013  Kipp Cannon, Ed Maros
+ * Copyright (C) 2011--2014  Kipp Cannon, Ed Maros
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -72,6 +72,7 @@
 
 #include <gstlal/gstlal.h>
 #include <gstlal/gstlal_debug.h>
+#include <gstlal/gstlal_frhistory.h>
 #include <gstlal/gstlal_tags.h>
 #include <framecpp_channeldemux.h>
 #include <gstfrpad.h>
@@ -171,11 +172,7 @@ static gboolean is_requested_channel(GstFrameCPPChannelDemux *element, const cha
  */
 
 
-#ifdef HAVE_FRAMECPP_2x
 static GstCaps *FrVect_get_caps(LDASTools::AL::SharedPtr<FrameCPP::FrVect> vect, gint *rate, guint *unit_size)
-#else
-static GstCaps *FrVect_get_caps(General::SharedPtr < FrameCPP::FrVect > vect, gint *rate, guint *unit_size)
-#endif
 {
 	GstCaps *caps;
 	gint width = vect->GetTypeSize() * 8;
@@ -248,11 +245,7 @@ static GstCaps *FrVect_get_caps(General::SharedPtr < FrameCPP::FrVect > vect, gi
  */
 
 
-#ifdef HAVE_FRAMECPP_2x
 static GstBuffer *FrVect_to_GstBuffer(LDASTools::AL::SharedPtr<FrameCPP::FrVect> vect, GstClockTime timestamp, guint64 offset, gint *rate, guint *unit_size)
-#else
-static GstBuffer *FrVect_to_GstBuffer(General::SharedPtr < FrameCPP::FrVect > vect, GstClockTime timestamp, guint64 offset, gint *rate, guint *unit_size)
-#endif
 {
 	GstBuffer *buffer;
 
@@ -294,7 +287,7 @@ static GstBuffer *FrVect_to_GstBuffer(General::SharedPtr < FrameCPP::FrVect > ve
 	 * set timestamp and duration
 	 */
 
-	GST_BUFFER_TIMESTAMP(buffer) = timestamp;
+	GST_BUFFER_TIMESTAMP(buffer) = timestamp + (GstClockTime) round(vect->GetDim(0).GetStartX() * GST_SECOND);
 	GST_BUFFER_DURATION(buffer) = gst_util_uint64_scale_int(vect->GetNData(), GST_SECOND, *rate);
 	GST_BUFFER_OFFSET(buffer) = offset;
 	GST_BUFFER_OFFSET_END(buffer) = offset + vect->GetNData();
@@ -575,11 +568,7 @@ static gboolean src_pad_do_pending_events(GstFrameCPPChannelDemux *element, GstP
  */
 
 
-#ifdef HAVE_FRAMECPP_2x
 static GstFlowReturn frvect_to_buffer_and_push(GstFrameCPPChannelDemux *element, GstPad *pad, LDASTools::AL::SharedPtr<FrameCPP::FrVect> vect, GstClockTime timestamp)
-#else
-static GstFlowReturn frvect_to_buffer_and_push(GstFrameCPPChannelDemux *element, GstPad *pad, General::SharedPtr < FrameCPP::FrVect > vect, GstClockTime timestamp)
-#endif
 {
 	struct pad_state *pad_state = (struct pad_state *) gst_pad_get_element_private(pad);
 	GstBuffer *buffer;
@@ -938,6 +927,10 @@ static GstFlowReturn chain(GstPad *pad, GstBuffer *inbuf)
 
 		GST_LOG_OBJECT(element, "found version %u frame file generated by %s library revision %d", element->frame_format_version, element->frame_library_name, element->frame_library_version);
 
+		/*
+		 * loop over frames
+		 */
+
 		while(1) {
 			FrameCPP::IFrameStream::frame_h_type frame;
 			try {
@@ -949,8 +942,6 @@ static GstFlowReturn chain(GstPad *pad, GstBuffer *inbuf)
 
 			GstClockTime frame_timestamp = 1000000000L * frame->GetGTime().GetSeconds() + frame->GetGTime().GetNanoseconds();
 
-			GST_LOG_OBJECT(element, "found frame %d at %" GST_TIME_SECONDS_FORMAT, ifs.GetCurrentFrameOffset(), GST_TIME_SECONDS_ARGS(frame_timestamp));
-
 			/*
 			 * update element properties
 			 */
@@ -968,6 +959,8 @@ static GstFlowReturn chain(GstPad *pad, GstBuffer *inbuf)
 			element->frame_number = frame->GetFrame();
 			g_object_notify(G_OBJECT(element), "frame-number");
 
+			GST_LOG_OBJECT(element, "frame index %d: #%d at %" GST_TIME_SECONDS_FORMAT, ifs.GetCurrentFrameOffset(), element->frame_number, GST_TIME_SECONDS_ARGS(frame_timestamp));
+
 			/*
 			 * populate tags from frame metadata.  the tags
 			 * pushed out a source pad are taken from the pad's
@@ -986,6 +979,31 @@ static GstFlowReturn chain(GstPad *pad, GstBuffer *inbuf)
 			g_free(container_format);
 			}
 
+			/*
+			 * retrieve frame-level FrHistory objects
+			 */
+
+			g_value_array_free(element->frame_history);
+			element->frame_history = g_value_array_new(0);
+			for(FrameCPP::FrameH::history_iterator current = frame->RefHistory().begin(), last = frame->RefHistory().end(); current != last; current++) {
+#ifndef G_VALUE_INIT
+				GValue value = {0};	/* FIXME:  remove when we can rely on glib >= 2.30 */
+#else
+				GValue value = G_VALUE_INIT;
+#endif
+				gchar *str;
+				GstLALFrHistory *history = gstlal_frhistory_new((*current)->GetName().c_str());
+				gstlal_frhistory_set_timestamp(history, (*current)->GetTime() * GST_SECOND);
+				gstlal_frhistory_set_comment(history, (*current)->GetComment().c_str());
+				str = gstlal_frhistory_to_string(history);
+				GST_LOG_OBJECT(element, "FrHistory: %s", str);
+				g_free(str);
+				g_value_init(&value, GSTLAL_FRHISTORY_TYPE);
+				g_value_take_boxed(&value, history);
+				g_value_array_append(element->frame_history, &value);
+			}
+			g_object_notify(G_OBJECT(element), "frame-history");
+
 			/*
 			 * process ADC data
 			 */
@@ -1013,12 +1031,14 @@ static GstFlowReturn chain(GstPad *pad, GstBuffer *inbuf)
 						continue;
 					}
 					srcpad = get_src_pad(element, name, GST_FRPAD_TYPE_FRADCDATA, &pads_added);
-					/* FIXME:  bias, slope, units */
+					/* FIXME:  units */
 					g_object_set(srcpad,
 						"comment", (*current)->GetComment().c_str(),
 						"channel-group", (*current)->GetChannelGroup(),
 						"channel-number", (*current)->GetChannelNumber(),
 						"nbits", (*current)->GetNBits(),	/* FIXME:  set depth in caps */
+						"bias", (*current)->GetBias(),
+						"slope", (*current)->GetSlope(),
 						NULL
 					);
 					if(!gst_pad_is_linked(srcpad)) {
@@ -1277,7 +1297,8 @@ enum property {
 	ARG_FRAME_LIBRARY_NAME,
 	ARG_FRAME_NAME,
 	ARG_FRAME_RUN,
-	ARG_FRAME_NUMBER
+	ARG_FRAME_NUMBER,
+	ARG_FRAME_HISTORY
 };
 
 
@@ -1377,6 +1398,10 @@ static void get_property(GObject *object, guint id, GValue *value, GParamSpec *p
 		g_value_set_uint(value, element->frame_number);
 		break;
 
+	case ARG_FRAME_HISTORY:
+		g_value_set_boxed(value, element->frame_history);
+		break;
+
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, id, pspec);
 		break;
@@ -1406,6 +1431,8 @@ static void finalize(GObject * object)
 	element->frame_library_name = NULL;
 	g_free(element->frame_name);
 	element->frame_name = NULL;
+	g_value_array_free(element->frame_history);
+	element->frame_history = NULL;
 
 	G_OBJECT_CLASS(parent_class)->finalize(object);
 }
@@ -1615,6 +1642,23 @@ static void framecpp_channeldemux_class_init(GstFrameCPPChannelDemuxClass *klass
 			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
 		)
 	);
+	g_object_class_install_property(
+		gobject_class,
+		ARG_FRAME_HISTORY,
+		g_param_spec_value_array(
+			"frame-history",
+			"Frame-level history list",
+			"List of GstFrHistory objects.",
+			g_param_spec_boxed(
+				"history",
+				"History entry",
+				"GstFrHistory object.",
+				GSTLAL_FRHISTORY_TYPE,
+				(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
+			),
+			(GParamFlags) (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS)
+		)
+	);
 }
 
 
@@ -1645,4 +1689,5 @@ static void framecpp_channeldemux_init(GstFrameCPPChannelDemux *element, GstFram
 	element->frame_name = g_strdup(DEFAULT_FRAME_NAME);
 	element->frame_run = DEFAULT_FRAME_RUN;
 	element->frame_number = DEFAULT_FRAME_NUMBER;
+	element->frame_history = g_value_array_new(0);
 }
